from mincly.io.common import (
    Io as _Io,
    Writer as _Writer,
    Reader as _Reader,
    HybridIo as _HybridIo,
)
from mincly.io.standard import (
    StandardTerminalReader as _StandardTerminalReader,
    StandardTerminalWriter as _StandardTerminalWriter,
)
import typing as _t
from ..io import Io as _Io, Reader as _Reader, Writer as _Writer
from ..screen import Screen as _Screen
from capyutils import Result as _Result

_T = _t.TypeVar("_T")


class Menu:
    """
    This class represents a menu system for user interaction.
    It behaves using a screen stack.
    """

    def __init__(
        self,
        input_output: _t.Optional[_Io] = None,
        input: _t.Optional[_Reader] = None,
        output: _t.Optional[_Writer] = None,
    ) -> None:
        """
        - `input_output`: An optional `Io` object that provides both input and
                output functionalities. If provided, `input` and `output` are
                ignored

        - `input`: An optional `Reader` object for handling user input.
                If not provided and `input_output` is also not provided, a
                `StandardTerminalReader` will be used by default. If
                `input_output` is provided, this argument is ignored

        - `output`: An optional `Writer` object for handling output to the user.
                If not provided and `input_output` is also not provided, a
                `StandardTerminalWriter` will be used by default. If
                `input_output` is provided, this argument is ignored
        """
        self._io: _Io
        if input_output is not None:
            self._io = input_output
        else:
            input_or_standard = (
                input if input is not None else _StandardTerminalReader()
            )
            output_or_standard = (
                output if output is not None else _StandardTerminalWriter()
            )
            self._io = _HybridIo(input_or_standard, output_or_standard)
        self._screen_stack: _t.List[_Screen[_t.Any]] = []
        self._last_input: _t.Any = None

    def show(self, message: str):
        """
        Shows a message without expecting any input from the user
        """
        self._io.print_overwrite(message)

    def push(self, screen: _Screen[_T]):
        """
        Adds the provided screen, which is added to this Menu instance's stack.

        Screen will only be displayed when calling `get_input()`
        """
        self._screen_stack.append(screen)

    def pop(self):
        """
        Pops current screen from the stack and returns to last found screen. If
        the current screen is the only screen in the stack, this method will
        raise an exception.
        """

        if len(self._screen_stack) < 2:
            raise RuntimeError(
                "Can't navigate back when current screen is root screen."
            )
        self._screen_stack.pop()

    def get_input(self) -> _t.Any:
        """
        Blocks until user provides a valid input. 'Valid input' is defined by
        the `Screen` class that is on the top of the screen stack.
        """
        self.prompt(self._screen_stack[-1])

    def get_last_input(self) -> _t.Any:
        return self._last_input

    def prompt(self, screen: _Screen[_T]) -> _T:
        """
        Displays the provided screen. Blocks until user provides a valid input.

        Provided screen is not added to screen stack and will not be stored by
        this `Menu` instance.
        """
        screen_result: _t.Union[_Result[_T], None] = None

        while screen_result is None or screen_result.is_err():

            if isinstance(screen_result, _Result):
                screen_result.is_err()
            self._io.print_overwrite(screen.get_display_string())

            if screen_result is not None and screen_result.is_err():
                self._io.write(f"<ERROR>: {screen_result.unwrap_err()}\n")

            user_input = self._io.read()

            screen_result = screen.process_input(user_input)
        result_value = screen_result.unwrap()
        self._last_input = result_value
        return result_value

    def current_screen_name(self) -> _t.Union[str, None]:
        return self._screen_stack[-1].name



from .common import (
    Reader as Reader,
    SupportsWrite as SupportsWrite,
    Writer as Writer,
    Io as Io,
    HybridIo as HybridIo,
)
from .ansi import (
    AnsiTerminalIo as AnsiTerminalIo,
    AnsiTerminalWriter as AnsiTerminalWriter,
)
from .standard import (
    StandardTerminalReader as StandardTerminalReader,
    StandardTerminalWriter as StandardTerminalWriter,
    StandardTerminalIo as StandardTerminalIo,
)



import os as _os
from mincly.io.common import Io as _Io, Writer as _Writer


class AnsiTerminalWriter(_Writer):
    """For ANSI compliant terminals, everything outputed to this terminal is
    stored and can be cleared and overwritten at any type. For it to work
    properly, this must be the only source of printed content."""

    def __init__(
        self, add_newline_to_prints: bool = False, always_flush: bool = True
    ) -> None:
        self._last_printed_content: str = ""
        self._print_end = "\n" if add_newline_to_prints else ""
        self._print_flush = always_flush

    def write(self, message: str):
        """Prints message to terminal. Avoid using ANSI control sequence
        characters in `value`"""
        self._last_printed_content += message + self._print_end
        print(message, end=self._print_end, flush=self._print_flush)

    def clear_last_n_lines(self, n: int):
        """Clears last `n` lines in terminal. Does not change internal printed
        content, you may use this method to rectify content printed outside this
        class."""
        if n < 1:
            return
        n = min(n, _os.get_terminal_size().lines)
        clear_string = "\033[2K"
        if n > 1:
            clear_string += "\033[A\033[2K" * (n - 1)
        print(clear_string, end="", flush=True)

    def clear(self):
        """Clears all content that this class printed. Does not account for
        printed content from other sources"""
        if self._last_printed_content == "":
            return
        terminal_width = _os.get_terminal_size().columns
        printed_lines = self._last_printed_content.split("\n")
        number_of_lines_in_terminal = 0
        for printed_line in printed_lines:
            number_of_lines_in_terminal += 1
            remaining_string = printed_line
            while len(remaining_string) > terminal_width:
                number_of_lines_in_terminal += 1
                remaining_string = remaining_string[terminal_width:]
        self.clear_last_n_lines(number_of_lines_in_terminal)
        self._last_printed_content = ""


class AnsiTerminalIo(AnsiTerminalWriter, _Io):
    """For ANSI compliant terminals, everything outputed to this terminal is
    stored and can be cleared and overwritten at any type. For it to work
    properly, this must be the only source of printed content.

    `read()` method is counted towards printed content for the purposes of
    `clear`ing the terminal."""

    def read(self) -> str:
        """Retrieves input using Python's builtin `input` method. Takes into
        account the user's input and newline character (ENTER) for the next call
        to `clear()`."""
        user_input = input()
        self._last_printed_content += user_input + "\n"
        return user_input



import typing as _t
from .common import Screen as _Screen
from capyutils import Result as _Result, Ok as _Ok, Err as _Err

_T = _t.TypeVar("_T")


class SelectScreen(_Screen[_T]):
    """
    Screen for user to select from one of many custom options.

    - `numbered_options` is any iterable of tuples of `(description: str, value:
      T)`. The user will be shown all numbered options in order, with a number
      indicating which number the user should provide to choose that option
      (starting with 1). `value` will be the returned value, should the user
      pick the associated option.

    - `keyword_options` works the same way as `numbered_options`, but instead of
      showing the user multiple items ordered from 1 onwards, they are instead
      associated with the given dictionary's keys. (e.g. `{"Open": ("opens
      something": "open_file")}` would return `"open_file"` if the user typed in
      `"Open"`)

    - `header` displays an header message above the selection listing

    - `name` the name of the screen, which acts as an identifier, should it be
      useful for the program's logic
    """

    def __init__(
        self,
        numbered_options: _t.Optional[_t.Sequence[_t.Tuple[str, _T]]] = None,
        keyword_options: _t.Optional[_t.Dict[str, _t.Tuple[str, _T]]] = None,
        header: str = "Pick an option:",
        name: _t.Union[str, None] = None,
    ) -> None:
        super().__init__(name)
        if numbered_options is None and keyword_options is None:
            raise ValueError(
                "SelectScreen can't be built with no options. At least one of numbered_options or keyword_options needs to not be 'None'"
            )
        self.numbered_options = numbered_options or tuple()
        self.keyword_options = keyword_options or dict()
        self.header = header if header.endswith("\n") else header + "\n"

    def process_input(self, user_input: str) -> _Result[_T]:
        if len(user_input) < 1:
            return _Err("Empty input")

        if user_input.isdecimal():
            nth_option = int(user_input) - 1
            if nth_option < 0 or nth_option >= len(self.numbered_options):
                return _Err(f"Invalid numbered option '{user_input}'")
            _, option = self.numbered_options[nth_option]
            return _Ok(option)

        _, option = self.keyword_options.get(user_input, ("", None))
        if option is None:
            return _Err(f"Invalid keyword option '{user_input}'")

        return _Ok(option)

    def get_display_string(self) -> str:
        display_string = f"{self.header}"

        for nth, (option_description, _) in enumerate(self.numbered_options, start=1):
            display_string += f" {nth} - {option_description}\n"

        if len(self.numbered_options) > 0:
            display_string += "\n"

        for key, (option_description, _) in self.keyword_options.items():
            display_string += f" {key} - {option_description}\n"

        if len(self.keyword_options) > 0:
            display_string += "\n"

        return display_string



from abc import ABC as _Abstract, abstractmethod as abstract
import typing as _t
from capyutils import Result as _Result

_T = _t.TypeVar("_T")


class Screen(_Abstract, _t.Generic[_T]):
    def __init__(self, screen_name: _t.Union[str, None] = None) -> None:
        self.name = screen_name

    def __str__(self) -> str:
        return f"{self.__class__.__name__}('{self.name if self.name is not None else '<NoName>'}')"

    @abstract
    def process_input(self, user_input: str) -> _Result[_T]:
        """
        React to `user_input` and give back a Result that represents the
        validity of the given input. `Result.Err` means input is invalid and its
        not this class's responsability what to do further. `Result.Ok` means
        input was valid and the wrapped value contains some sort of response
        that will only make sense for a concrete screen (or `None`)
        """

    @abstract
    def get_display_string(self) -> str:
        """
        Returns a string of what should be displayed for this screen
        """
